@@ -12,7 +12,6 @@ import shutil
 
 # Do not remove unused imports. They are used in the jinja file probably
 from pathlib import Path
-import os
 import stat
 import logging
 import enum
@@ -25,7 +24,7 @@ class KWARGS(dict):
     pass
 
 
-_app_list = []
+_function_dict = {}
 
 
 env = Environment(
@@ -69,9 +68,18 @@ class ExecutableDirectory:
     in the cluster (i.e. is located on a shared filesystem).
     """
 
-    def __init__(self, dirname, runsh="run.sh", _execution_resources=None) -> None:
+    def __init__(self, dirname, runsh=None, _execution_resources=None) -> None:
         self.dirname = dirname
-        self.runsh = runsh
+        if runsh is None:
+            runsh_files = glob(os.path.join(dirname, "run_*.sh"))
+            if len(runsh_files) == 0:
+                raise RuntimeError(f"No run script found in {dirname}")
+            runsh_files.sort(key=os.path.getmtime)
+            self.runsh = os.path.basename(runsh_files[-1])
+        else:
+            self.runsh = runsh
+
+        self.sentinel_dir = os.path.join(dirname, self.runsh.replace(".sh", ""))
         self.execution_resources = _execution_resources
         self.slurmq = None
 
@@ -93,13 +101,20 @@ class ExecutableDirectory:
             return self._pmws_status()
 
         # If the container reported completion or error
-        if os.path.exists(os.path.join(self.dirname, ".pm.done")):
+        # TODO: Remove the checks in dirname once all apps have been updated to write in the sentinel_dir
+        if os.path.exists(
+            os.path.join(self.sentinel_dir, ".pm.done")
+        ) or os.path.exists(os.path.join(self.dirname, ".pm.done")):
             return JobStatus.COMPLETED
-        elif os.path.exists(os.path.join(self.dirname, ".pm.err")):
+        elif os.path.exists(
+            os.path.join(self.sentinel_dir, ".pm.err")
+        ) or os.path.exists(os.path.join(self.dirname, ".pm.err")):
             return JobStatus.ERROR
 
         # If the container is reporting aliveness
-        heartbeat = os.path.join(self.dirname, ".pm.alive")
+        heartbeat = os.path.join(self.sentinel_dir, ".pm.alive")
+        if not os.path.exists(heartbeat):
+            heartbeat = os.path.join(self.dirname, ".pm.alive")
         if os.path.exists(heartbeat):
             with open(heartbeat, "r") as f:
                 timestamp_str = f.read().strip()
@@ -138,13 +153,13 @@ class ExecutableDirectory:
             return mapping[info[list(info.keys())[0]]["state"]]
 
         # Alternatively check if expected outputs exist
-        outputs = os.path.join(self.dirname, "expected_outputs.json")
+        outputs = os.path.join(self.sentinel_dir, "expected_outputs.json")
         if os.path.exists(outputs):
             with open(outputs, "r") as f:
                 outputs = json.load(f)
 
             for outf in outputs:
-                if not os.path.exists(outf):
+                if len(glob(os.path.join(self.dirname, outf))) == 0:
                     return JobStatus.RUNNING
             else:
                 return JobStatus.COMPLETED
@@ -154,7 +169,7 @@ class ExecutableDirectory:
         )
         return JobStatus.WAITING_INFO
 
-    def run(self, queue=None, **kwargs):
+    def run(self, queue=None, verbose=True, **kwargs):
         """Execute the directory locally
 
         If no queue is specified it will run the job locally.
@@ -183,7 +198,21 @@ class ExecutableDirectory:
         if queue is None:
             import subprocess
 
-            subprocess.call(["bash", self.runsh], cwd=self.dirname)
+            logfile = os.path.join(self.dirname, self.runsh.replace(".sh", ".log"))
+
+            with open(logfile, "w") as logfile:
+                process = subprocess.Popen(
+                    ["bash", self.runsh],
+                    cwd=self.dirname,
+                    stdout=subprocess.PIPE,
+                    stderr=subprocess.STDOUT,
+                    text=True,
+                )
+                for line in process.stdout:
+                    if verbose:
+                        print(line, end="")  # Write to terminal
+                    logfile.write(line)  # Write to file
+                process.wait()
         else:
             if queue.lower() == "slurm":
                 self.slurm(**kwargs)
@@ -318,7 +347,7 @@ class ExecutableDirectory:
         import requests
         import json
 
-        with open(os.path.join(self.dirname, ".manifest.json"), "r") as f:
+        with open(os.path.join(self.sentinel_dir, ".manifest.json"), "r") as f:
             manifest = json.load(f)
 
         endpoint = PM_APP_ROOT
@@ -595,7 +624,7 @@ def _write_inputs(write_dir, inputdir, arguments, manifest, function=None):
     os.makedirs(inputdir, exist_ok=True)
 
     if "outputs" in manifest:
-        with open(os.path.join(write_dir, "expected_outputs.json"), "w") as f:
+        with open(os.path.join(inputdir, "expected_outputs.json"), "w") as f:
             json.dump(manifest["outputs"], f)
 
     # Validate arg types and copy Path arguments to folder
@@ -729,7 +758,7 @@ def _write_inputs(write_dir, inputdir, arguments, manifest, function=None):
     with open(os.path.join(inputdir, "original_paths.json"), "w") as f:
         json.dump(original_paths, f, indent=4)
 
-    with open(os.path.join(write_dir, ".manifest.json"), "w") as f:
+    with open(os.path.join(inputdir, ".manifest.json"), "w") as f:
         json.dump(manifest, f, indent=4)
 
 
@@ -764,6 +793,15 @@ def _docs_from_manifest(manifest, appname, appdir):
     else:
         missing.append("outputs")
 
+    if "resources" in manifest and manifest["resources"] is not None:
+        docs.append("")
+        docs.append("Note")
+        docs.append("----")
+        docs.append("Minimum job requirements::")
+        docs.append("")
+        for key, val in manifest["resources"].items():
+            docs.append(f"    {key}: {val}")
+
     if "examples" in manifest:
         docs.append("")
         docs.append("Examples")
@@ -820,11 +858,6 @@ def _docs_from_manifest(manifest, appname, appdir):
         for fname in tutorials:
             docs.append(f"    - {fname}")
 
-    if "name" not in manifest:
-        missing.append("name")
-    else:
-        appname = manifest["name"]
-
     if len(missing):
         logger.warning(f"{appname} manifest is missing fields: {', '.join(missing)}")
     return docs
@@ -875,10 +908,62 @@ def _args_from_manifest(func_args):
     return args
 
 
+def __create_function(
+    function,
+    function_idx,
+    function_name,
+    function_args,
+    function_docs,
+    function_resources,
+    module_path,
+    new_mode,
+    run_sh,
+):
+    template = env.get_template("func.py.jinja")
+    code = template.render(
+        function=function,
+        function_idx=function_idx,
+        function_name=function_name,
+        function_args=function_args,
+        function_docs=function_docs,
+        function_resources=function_resources,
+        module_path=module_path,
+        new_mode=new_mode,
+        run_sh=run_sh,
+    )
+    local_ns = {}
+    exec(code, globals().copy(), local_ns)
+    return local_ns[function_name]
+
+
+def __ensure_module_path(module_path):
+    """
+    Ensures that all parts of the dotted module path exist in sys.modules
+    and returns the final module object.
+    """
+    import sys
+    import types
+
+    parts = module_path.split(".")
+    full_path = ""
+    parent = None
+
+    for part in parts:
+        full_path = f"{full_path}.{part}" if full_path else part
+        if full_path not in sys.modules:
+            mod = types.ModuleType(full_path)
+            sys.modules[full_path] = mod
+            if parent:
+                setattr(parent, part, mod)
+        parent = sys.modules[full_path]
+
+    return sys.modules[full_path]
+
+
 def _manifest_to_func(appname, app_versions, latest):
     from copy import deepcopy
+    import sys
 
-    app_info = {}
     for version in app_versions:
         manifest = app_versions[version]["manifest"]
         new_mode = True
@@ -908,56 +993,96 @@ def _manifest_to_func(appname, app_versions, latest):
                 import traceback
 
                 logger.error(
-                    f"Failed to parse manifest for app {manifest['name']} version {version} due to error:\n{traceback.format_exc()}"
+                    f"Failed to parse manifest for app {appname} version {version} due to error:\n{traceback.format_exc()}"
                 )
                 continue
 
-        app_info[version] = {
-            "run.sh": app_versions[version]["run.sh"],
-            "manifest": manifest,
-            "appdir": app_versions[version]["appdir"],
-            "functions": [],
-            "main_function": {
-                "args": [],
-                "docs": ["Main function not implemented. See other methods!"],
-            },
-            "tests": {},
-            "new_mode": new_mode,
-        }
-        try:
-            for func in manifest["functions"]:
-                func = deepcopy(func)
-                func["name"] = func["function"].split(".")[-1]
-                func["args"] = _args_from_manifest(func["params"])
-                func["docs"] = _docs_from_manifest(
-                    func, appname, app_versions[version]["appdir"]
+        module_path = f"playmolecule.apps.{appname}.{version}"
+        module = __ensure_module_path(module_path)
+
+        setattr(module, "__manifest__", deepcopy(manifest))
+
+        app_files = _get_app_files(
+            os.path.join(app_versions[version]["appdir"], "files"), appname, version
+        )
+        setattr(module, "files", app_files)
+
+        artifact_dict = manifest.get("artifacts", manifest.get("datasets", {}))
+        app_artifacts = None
+        if len(artifact_dict):
+            app_artifacts = _Artifacts(artifact_dict, app_files)
+            setattr(module, "artifacts", app_artifacts)
+            setattr(module, "datasets", app_artifacts)
+
+        func_names = []
+        for idx, func_mani in enumerate(manifest["functions"]):
+            try:
+                func_name = func_mani["function"].split(".")[-1]
+                if func_name == "main":
+                    func_name = appname
+                _func = __create_function(
+                    function=func_mani["function"],
+                    function_idx=idx,
+                    function_name=func_name,
+                    function_args=_args_from_manifest(func_mani["params"]),
+                    function_docs=_docs_from_manifest(
+                        func_mani, appname, app_versions[version]["appdir"]
+                    ),
+                    function_resources=func_mani.get("resources", None),
+                    module_path=module_path,
+                    new_mode=new_mode,
+                    run_sh=app_versions[version]["run.sh"],
                 )
-                app_info[version]["functions"].append(func)
-                if func["name"] == "main":
-                    app_info[version]["main_function"] = func
+                tests = _Tests(func_mani["tests"], _func, app_files, app_artifacts)
+                # Add some metadata to the function object
+                _func.tests = tests
+                _func._name = func_name
+                _func.__manifest__ = deepcopy(func_mani)
+                func_names.append(func_name)
+                setattr(module, func_name, _func)
+                _function_dict[f"{module_path}.{func_name}"] = _func
+            except Exception:
+                import traceback
 
-                if "tests" in func:
-                    for test in func["tests"]:
-                        app_info[version]["tests"][test] = func["tests"][test]
-                        app_info[version]["tests"][test]["function"] = func["name"]
-        except Exception:
-            import traceback
+                logger.error(
+                    f"Failed to parse manifest for app {appname} version {version} with error: {traceback.format_exc()}"
+                )
 
-            logger.error(
-                f"Failed to parse manifest for app {manifest['name']} version {version} with error: {traceback.format_exc()}"
-            )
+        # Link the latest version of the app to the top level module
+        parent_module = sys.modules[f"playmolecule.apps.{appname}"]
+        latest_module = sys.modules[f"playmolecule.apps.{appname}.{latest}"]
+        for symbol in func_names + [
+            "artifacts",
+            "datasets",
+            "files",
+            "tests",
+            "__manifest__",
+        ]:
+            if (
+                symbol not in parent_module.__dict__
+                and symbol in latest_module.__dict__
+            ):
+                setattr(
+                    sys.modules[f"playmolecule.apps.{appname}"],
+                    symbol,
+                    getattr(
+                        sys.modules.get(f"playmolecule.apps.{appname}.{latest}"), symbol
+                    ),
+                )
 
-    if len(app_info) == 0:
-        return
 
-    template = env.get_template("func.py.jinja")
-    fstring = template.render(
-        app_info=app_info,
-        app_name=appname,
-        class_name=appname.capitalize(),
-        latest=latest,
-    )
-    exec(fstring)
+def _check_folder_validity(root_dir, app_d):
+    import re
+
+    folder_path = os.path.relpath(app_d, root_dir)
+
+    # Check if the folder consists only of lowercase letters, numbers, underscores or path separators
+    if not re.match(r"^[a-z0-9_/\\]+$", folder_path):
+        logger.warning(
+            f'Path {app_d} has invalid characters in the part: "{folder_path}". Only lowercase letters, numbers and underscores are allowed. Please fix the path. Skipping...'
+        )
+        return False
+    return True
 
 
 def _set_root(root_dir):
@@ -977,12 +1102,17 @@ def _set_root(root_dir):
 
     for app_d in natsorted(glob(os.path.join(root_dir, "apps", "*", ""))):
         appname = os.path.basename(os.path.abspath(app_d))
+        if not _check_folder_validity(root_dir, app_d):
+            continue
+
         versions = glob(os.path.join(app_d, "*"))
         versions.sort(key=lambda s: natsorted(os.path.basename(s)))
 
         app_versions = {}
         for vv in versions:
             vname = os.path.basename(vv)
+            if not _check_folder_validity(root_dir, vv):
+                continue
 
             jf = glob(os.path.join(vv, "*.json"))
             if len(jf) > 1:
@@ -1008,7 +1138,7 @@ def _set_root(root_dir):
         with open(dsjson) as f:
             manifest = json.load(f)
         files = _get_app_files(dsdir)
-        playmolecule.datasets = _Artifacts(manifest, files)
+        playmolecule.datasets = _Artifacts(manifest["datasets"], files)
 
 
 def _setup_folders(root_dir):
